/* usadelndsoc
 *
 * Copyright © 2022 Pauli Virtanen
 *    @author Pauli Virtanen <pauli.t.virtanen@jyu.fi>
 *
 * SPDX-License-Identifier: AGPL-3.0-or-later
 */
#ifndef SIMPLEVECTOR_HPP_
#define SIMPLEVECTOR_HPP_

#include <vector>

namespace array {

template <class Scalar>
class Vector
{
public:
    typedef Scalar value_type;

private:
    static_assert(!std::is_same_v<Scalar,bool>, "Vector<bool> is not supported");
    std::vector<Scalar> storage_;
    Scalar *data_;
    size_t size_;

    bool own_data() const
        {
            return data_ == storage_.data();
        }

    void copy_storage()
        {
            if (own_data())
                return;

            storage_.resize(size_);
            for (size_t i = 0; i < size_; ++i)
                storage_[i] = data_[i];

            data_ = storage_.data();
            size_ = storage_.size();
        }

public:
    Vector()
        : storage_(), data_(storage_.data()), size_(storage_.size())
        {}

    Vector(size_t n)
        : storage_(n), data_(storage_.data()), size_(storage_.size())
        {}

    Vector(const std::initializer_list<Scalar>& values)
        : storage_(values), data_(storage_.data()), size_(storage_.size())
        {}

    Vector(const Vector& other)
        : storage_(other.storage_), data_(other.data_), size_(other.size_)
        {
            if (other.own_data()) {
                data_ = storage_.data();
                size_ = storage_.size();
            } else {
                copy_storage();
            }
        }

    Vector(Vector&& other)
        : storage_(std::move(other.storage_)), data_(other.data_), size_(other.size_)
        {
            other.data_ = other.storage_.data();
            other.size_ = other.storage_.size();
        }

    Vector(Scalar *data, size_t size)
        : storage_(), data_(data), size_(size)
        {}

    Vector& operator=(const Vector& other)
        {
            if (other.own_data()) {
                storage_ = other.storage_;
                data_ = storage_.data();
                size_ = storage_.size();
            } else {
                data_ = other.data_;
                size_ = other.size_;
                copy_storage();
            }
            return *this;
        }

    Vector& operator=(Vector&& other)
        {
            if (other.own_data()) {
                storage_ = std::move(other.storage_);
                data_ = storage_.data();
                size_ = storage_.size();
            } else {
                storage_ = std::move(other.storage_);
                data_ = other.data_;
                size_ = other.size_;
                copy_storage();
            }
            other.data_ = other.storage_.data();
            other.size_ = other.storage_.size();
            return *this;
        }

    void swap(Vector& other)
        {
            storage_.swap(other.storage_);
            std::swap(data_, other.data_);
            std::swap(size_, other.size_);
        }

    Scalar *data() const
        {
            return data_;
        }

    size_t size() const
        {
            return size_;
        }

    void resize(size_t n)
        {
            copy_storage();

            storage_.resize(n);
            data_ = storage_.data();
            size_ = storage_.size();
        }

    Scalar& operator[](size_t i)
        {
            return data_[i];
        }

    const Scalar& operator[](size_t i) const
        {
            return data_[i];
        }
};

}

#endif // SIMPLEVECTOR_HPP_
